/**
* conf_cl.c - Command line interface Client for the Linksys WRT54GS configuration server
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <termios.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>

#include "conffile.h"

#define ACTION_GET 10
#define ACTION_SET 20
#define ACTION_UNSET 30
#define ACTION_ACK 40
#define ACTION_ERR 50
#define ACTION_AUTH 60
#define ACTION_AUTH_OK 70
#define ACTION_AUTH_ERR 80

#define DEF_SRV_PORT 5555

typedef struct msg_t {
	unsigned short mtype;
	unsigned char *mtext;
} MSG;

unsigned char *make_msg(unsigned short action, unsigned char *var, unsigned char *val) {
	unsigned char *text;
	int len = 0;
	int clen = 0;
	unsigned char strcl[10];
	memset(strcl, 0, 10);

	len = 3 + strlen("\r\nContent-length: ");
	switch (action) {
		case ACTION_GET:
			clen += strlen("VAR: ") + strlen(var);
			sprintf(strcl, "%d", clen);
			len += strlen(strcl) + clen + 3; //add length of string repr of clen (+3: 2 for the \r\n and 1 xtra null byte - string end
			text = (unsigned char*)malloc(len);
			memset(text, 0, len);
			sprintf(text, "GET\r\nContent-length: %d\r\nVAR: %s", clen, var);
			return text;

		case ACTION_SET:
			clen += strlen("VAR: ") + strlen(var) + 2 + strlen("VAL: ") + strlen(val);
			sprintf(strcl, "%d", clen);
			len += strlen(strcl) + clen + 3; //3: 2 for the crlf after "content-lengt" and one for the final NULL byte
			text = (unsigned char*)malloc(len);
			memset(text, 0, len);
			sprintf(text, "SET\r\nContent-length: %d\r\nVAR: %s\r\nVAL: %s", clen, var, val);
			return text;

		case ACTION_UNSET:
			clen += strlen("VAR: ") + strlen(var);
			sprintf(strcl, "%d", clen);
			len += strlen(strcl) + clen + 3;
			text = (unsigned char*)malloc(len);
			memset(text, 0, len);
			sprintf(text, "UST\r\nContent-length: %d\r\nVAR: %s", clen, var);
			return text;

		case ACTION_AUTH:
			clen += strlen("VAR: ") + strlen(var);
			sprintf(strcl, "%d", clen);
			len += strlen(strcl) + clen + 3;
			text = (unsigned char*)malloc(len);
			memset(text, 0, len);
			sprintf(text, "AUT\r\nContent-length: %d\r\nVAR: %s", clen, var);
			return text;

		default:
			return NULL;
	}
}

unsigned char *parse_msg(const unsigned char *text) {
	unsigned char text_copy[strlen(text)+1];
	memset(text_copy, 0, strlen(text) + 1);
	strcpy(text_copy, text);
	int clen = 0;
	unsigned char *pos = NULL;
	unsigned char *crlfpos = NULL;
	unsigned char strclen[10];
	memset(strclen, 0, 10);

	unsigned char *var = NULL;

	if (!strstr(text, "ACK\r\nContent-length: ") && !strstr(text, "ERR\r\nContent-length: ")) {
		return NULL;
	}

	pos = text_copy + 3 + strlen("\r\nContent-length: ");
	crlfpos = strstr(pos, "\r\n");
	if (!crlfpos) {
		return NULL;
	}
	strncpy(strclen, pos, crlfpos - pos);
	clen = atoi(strclen);
	pos = crlfpos + 2;

	if (strncasecmp(pos, "VAR: ", 5)) {
		return NULL;
	}
	pos += 5;

	var = (unsigned char*)malloc(clen - strlen("VAR: ") + 1);
	memset(var, 0, clen - strlen("VAR: ") + 1);
	strncpy(var, pos, clen - strlen("VAR: "));
	return var;
}

MSG *read_message(int sockfd) {
	//first read the first 4 bytes (message type) - check if its a conf msg, else exit
	unsigned char *text_message = NULL;
	MSG *retval = NULL;

	unsigned char *header = NULL; //content length header (this array is read from the socket)

	//message type string(read from the socket)
	unsigned char strtype[4];
	memset(strtype, 0, 4);

	unsigned char strcl[10]; //content length of message
	int clen = 0;
	int bytesread = 0;
	int totlen = 0; //total len of the text message
	int mtype = 0;
	int i = 0;

	//read message type (the first 3 bytes)
	bytesread = read(sockfd, strtype, 3);
	if (bytesread != -1) {
		//save message type
		if (!strncasecmp(strtype, "ACK", 3)) {
			mtype = ACTION_ACK;
		}
		else
		if (!strncasecmp(strtype, "ERR", 3)) {
			mtype = ACTION_ERR;
		}
		else
		if (!strncasecmp(strtype, "AOK", 3)) {
			mtype = ACTION_AUTH_OK;
		}
		else
		if (!strncasecmp(strtype, "AER", 3)) {
			mtype = ACTION_AUTH_ERR;
		}
		else {
			fprintf(stderr, "read_message: Unknown message type\n");
			return NULL;
		}

		if ( (mtype == ACTION_AUTH_OK) || (mtype == ACTION_AUTH_ERR) ) {
			//these messages have no body, just AOK or AER
			text_message = (unsigned char*)malloc(4);
			memset(text_message, 0, 4);
			strncpy(text_message, strtype, 3);
		}
		else {
			header = (unsigned char*)malloc(strlen("\r\nContent-length: ") + 1);
			memset(header, 0, strlen("\r\nContent-length: ") + 1);
			bytesread = read(sockfd, header, strlen("\r\nContent-length: "));
			if (bytesread != -1) {
				//first check header
				if (strcasecmp(header, "\r\nContent-length: ")) {
					//malformed header
					if (header) free(header);
					return NULL;
				}
				//read content length
				//actually, we read byte after byte until we find a crlf
				memset(strcl, 0, 10);
				i = 0;
				while ( !strstr(strcl, "\r\n") && (i < 10) ) {
					if (read(sockfd, &strcl[i], 1) == -1) {
						//error reading character
						if (header) free(header);
						return NULL;
					}
					i++;
				}
				if (!sscanf(strcl, "%d\r\n", &clen)) {
					//malformed clen
					if (header) free(header);
					return NULL;
				}

				//finally, read the remaining text message into "text_message" (which will eventually be returned)
				totlen = 3 + strlen(header) + strlen(strcl) + clen + 1; //total length of message ("3": mtype, "+1": trailing \0)
				text_message = (unsigned char*)malloc(totlen);
				memset(text_message, 0, totlen);

				//copy the fields that are already read
				sprintf(text_message, "%s%s%s", strtype, header, strcl);
				//"append" the remaining bytes to text_message (start writing from position text_message+sizeof(already read))
				bytesread = read(sockfd, text_message + strlen(text_message), clen);
				if (bytesread != clen) {
					printf("%d - %d - %s\n", bytesread, clen, text_message);
					//error reading data from socket: bytesread must be == with the calculated text message length
					if(text_message) free(text_message);
					if (header) free(header);
					return NULL;
				}
			}
			else {
				if (header) free(header);
				return NULL;
			}
		}
	}
	else {
		return NULL;
	}
	if(header) {
		free(header);
	}
	retval = (MSG*)malloc(sizeof(MSG));
	memset (retval, 0, sizeof(MSG));
	retval->mtype = mtype;
	retval->mtext = (unsigned char*)malloc(strlen(text_message) + 1);
	memset(retval->mtext, 0, strlen(text_message) + 1);
	strcpy(retval->mtext, text_message);
	if (text_message) free(text_message);

	return retval;
}

int auth_user(unsigned char *passwd, int sockfd) {
	int auth = 0; //0: !authenticated, 1: authenticated
	unsigned char *text = NULL;
	MSG *resp = NULL;

	//send auth request to srv
	text = make_msg(ACTION_AUTH, passwd, NULL);
	if (!text) {
		return 0;
	}

	if (!write(sockfd, text, strlen(text))) {
		printf("Error writing to socket\n");
		if (text) free(text);
		return 0;
	}
	if (text) free(text);

	//read response
	resp = read_message(sockfd);
	if (!resp) {
		return 0;
	}
	if (resp->mtype != ACTION_AUTH_OK) {
		auth = 0; //not authenticated
	}
	else {
		auth = 1; //authenticated
	}

	if (resp) {
		if (resp->mtext) {
			free(resp->mtext);
		}
		free(resp);
	}
	return auth;
}


void print_usage() {
	fprintf(stderr, "\nUsage: wrt_conf_client ");
	fprintf(stderr, "[-h/--help] ");
	fprintf(stderr, "[-srv/--server-addr] ");
	fprintf(stderr, "[-p/--port portnum] ");
	fprintf(stderr, "[-g/--get varname] ");
	fprintf(stderr, "[-s/--set varname] ");
	fprintf(stderr, "[-u/--unset varname] ");
	fprintf(stderr, "[-v/--value variable_value] ");
	fprintf(stderr, "[-f/--file path\n\n");

	fprintf(stderr, "\t-h --help\t\tDisplay this message\n");
	fprintf(stderr, "\t-srv --server-addr\tConfiguration server IP address\n");
	fprintf(stderr, "\t-p --port\t\tServer port (default: 5555)\n");
	fprintf(stderr, "\t-g --get\t\tGet variable value\n");
	fprintf(stderr, "\t-s --set\t\tSet variable value (must be used with -v/--value)\n");
	fprintf(stderr, "\t-u --unset\t\tUnset variable\n");
	fprintf(stderr, "\t-v --value\t\tSupply variable value (cannot be used with -f)\n");
	fprintf(stderr, "\t-f --file\t\tPath of the file where the variable value will be read from ");
	fprintf(stderr, "(cannot be used with -v)\n\n");
}

int main (int argc, char **argv) {
	struct sockaddr_in server_address;
	int sockfd = 0;
	MSG *msg = NULL;
	FILE * fp = NULL;
	int bytes_read = 0;
	int addr_len = sizeof(server_address);

	//message text
	unsigned char *text = NULL;

	//administrator password
	char password[80];
	char *nl = NULL;

	//configuration
	char str_server_addr[80];
	char get_var[80];
	char set_var[80];
	char unset_var[80];
	char val[8192];
	char value_file[256];

	int server_port = DEF_SRV_PORT;

	int help = 0;

	memset(&str_server_addr, 0, 80);
	memset(&get_var, 0, 80);
	memset(&set_var, 0, 80);
	memset(&unset_var, 0, 80);
	memset(&val, 0, 8192);
	memset(&value_file, 0, 256);

	CMD_ARG *cmd_line_args = NULL;

	unsigned char *valid_cmd_line_names[] ={
		"help",
		"server-addr",
		"port",
		"get",
		"set",
		"unset",
		"value",
		"file"
	};

	unsigned char* cmd_line_switches[] = {"h", "srv", "p", "g", "s", "u", "v", "f"};

	unsigned short cmd_line_var_types[] = {
		VTYPE_NOVAL,
		VTYPE_STRING,
		VTYPE_INT,
		VTYPE_STRING,
		VTYPE_STRING,
		VTYPE_STRING,
		VTYPE_STRING,
		VTYPE_STRING
	};

	void *cmd_line_variables[] = {
		&help,
		&str_server_addr,
		&server_port,
		&get_var,
		&set_var,
		&unset_var,
		&val,
		&value_file
	};

	cmd_line_args = make_valid_args(valid_cmd_line_names, cmd_line_switches, cmd_line_variables, cmd_line_var_types, 8);

	if (argc < 2) {
		print_usage();
		exit(1);
	}
	else {
		//parse command line arguments
		parse_cmd_line_args(argc, argv, cmd_line_args, 8);
		if (help) {
			//user gave the -h flag (probably among others...)
			print_usage();
			exit(0);
		}
	}

	//check whether server address has been supplied
	if (!strlen(str_server_addr)) {
		fprintf(stderr, "Access point address has not been specified. Please use the -srv option\n");
		print_usage();
		exit(1);
	}

	//check if an action has been requested
	if (!strlen(get_var) && !strlen(set_var) && !strlen(unset_var) ) {
		fprintf(stderr, "No action spefified\n");
		print_usage();
		exit(1);
	}

	//check mutually exclusive fields (set-unset)
	if ( strlen(set_var) && strlen(unset_var) ) {
		fprintf(stderr, "Mismatch: -s and -u options cannot be used together\n");
		print_usage();
		exit(1);
	}
	if ( strlen(set_var) && strlen(get_var) ) {
		fprintf(stderr, "Mismatch: -s and -g options cannot be used together\n");
		print_usage();
		exit(1);
	}
	if ( strlen(get_var) && strlen(unset_var) ) {
		fprintf(stderr, "Mismatch: -g and -u options cannot be used together\n");
		print_usage();
		exit(1);
	}
	if ( strlen(val) && strlen(value_file) ) {
		fprintf(stderr, "Mismatch: -v and -f options cannot be used together\n");
		print_usage();
		exit(1);
	}


	//check if the -s flag is used without the -v or the -f flag
	if ( strlen(set_var) && !strlen(val) && !strlen(value_file) ) {
		fprintf(stderr, "Missing value. Please add the -v or the -f option\n");
		print_usage();
		exit(1);
	}
	//end of parsing user input

	//check if the -f option has been supplied
	//if so, read the value from the file
	if (strlen(value_file)) {
		fp = fopen(value_file, "r");
		if (!fp) {
			fprintf(stderr, "Could not read variable value from file %s\n", value_file);
			print_usage();
			exit(1);
		}
		else {
			//read value from file
			int retval = 0;
			while ( (bytes_read < 8192) && !feof(fp) ) {
				if ( (fread(&val[bytes_read], 1, 1, fp) != 1) && ferror(fp) ) {
					fprintf(stderr, "Warning: Error while reading from file %s\n", value_file);
					break;
				}
				bytes_read++;
			}
		}
		if (fp) {
			fclose(fp);
		}
	}

	//prompt for user password
	fprintf(stderr, "Enter password:\n");

	struct termios old_term_settings;
	struct termios new_term_settings;

	//first, echo terminal echo off
	tcgetattr(0, &old_term_settings);
	new_term_settings = old_term_settings;
	new_term_settings.c_lflag &= (~ECHO);
	tcsetattr(0, TCSANOW, &new_term_settings);

	//then, read password
	memset(password, 0, 80);
	fgets(password, 80, stdin);

	//remove trailing newline
	nl = strchr(password, '\r');
	if (nl) {
		*nl = '\0';
	}
	nl = strchr(password, '\n');
	if (nl) {
		*nl = '\0';
	}

	//reset terminal
	tcsetattr(0, TCSANOW, &old_term_settings);

	memset((char  *)&server_address, 0, sizeof(server_address));
	server_address.sin_family = AF_INET;
	if (!inet_aton(str_server_addr, &(server_address.sin_addr))) {
		fprintf(stderr, "inet_aton failed\n");
		return 0;
	}
	server_address.sin_port = htons(server_port);

	sockfd = socket (PF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		fprintf(stderr, "socket creation failed\n");
		return 0;
	}

	connect(sockfd, (const struct sockaddr *)&server_address, addr_len);

	if (!auth_user(password, sockfd)) {
		fprintf(stderr, "Could not authenticate client. Please check your password\n");
		close(sockfd);
		exit(1);
	}
	else {
		fprintf(stderr, "Client authenticated\n");
	}

	if (strlen(set_var)) {
		//set variable
		text = make_msg(ACTION_SET, set_var, val);
		fprintf(stderr, "Requesting to set variable %s to value %s\n", set_var, val);
	}
	else
	if (strlen(get_var)) {
		//get var
		text = make_msg(ACTION_GET, get_var, NULL);
		fprintf(stderr, "Requesting to get variable %s\n", get_var);
	}
	else
	if (strlen(unset_var)) {
		//unset var
		text = make_msg(ACTION_UNSET, unset_var, NULL);
		fprintf(stderr, "Requesting to unset variable %s\n", unset_var);
	}
	else {
		fprintf(stderr, "Error: invalid message type\n");
		if (text) {
			free(text);
		}
		close(sockfd);
		exit(1);
	}

	if (!write(sockfd, text, strlen(text))) {
		fprintf(stderr, "Error writing to socket\n");
		if (text) {
			free(text);
		}
		exit(1);
	}

	msg = read_message(sockfd);
	if (!msg) {
		fprintf(stderr, "Error getting response from server\n");
	}
	else {
		if (msg->mtype == ACTION_ERR) {
			fprintf(stderr, "Negative server response - ERR message\n");
		}
		else {
			fprintf(stderr, "Action performed successfully\n");
		}
		
		if (strlen(get_var)) {
			fprintf(stderr, "Variable value: %s\n", parse_msg(msg->mtext));
		}
	}
	if (msg) {
		if (msg->mtext) {
			free(msg->mtext);
		}
		free(msg);
	}

	return 0;
}
